#!/usr/bin/env python
# coding=utf-8

"""
elzzur solves a Ruzzle board, listing all the valid words with their scores.

This is the main elzzur script, intended to be run from command line.
"""

from __future__ import absolute_import
from __future__ import print_function
import argparse
import os
import sys

from elzzur.board import Board 
from elzzur.languages import LANGUAGES
from elzzur.mtdictionary import MTDictionary
from elzzur.solver import Solver

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2016, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "0.0.1"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

DESCRIPTION = "elzzur solves a Ruzzle board"

ARGUMENTS = [
    {
        "long": "command",
        "short": None,
        "nargs": None,
        "type": str,
        "default": None,
        "help": "[cat|compile|demo|generate|languages|solve]"
    },
    {
        "long": "--language",
        "short": "-l",
        "nargs": "?",
        "type": str,
        "default": None,
        "help": "Language code (e.g., 'en')"
    },
    {
        "long": "--dictionary",
        "short": "-d",
        "nargs": "?",
        "type": str,
        "default": None,
        "help": "Path to the dictionary file"
    },
    {
        "long": "--board",
        "short": "-b",
        "nargs": "?",
        "type": str,
        "default": None,
        "help": "Path to the board file"
    },
    {
        "long": "--output",
        "short": "-o",
        "nargs": "?",
        "type": str,
        "default": None,
        "help": "Path to the output file"
    },
    {
        "long": "--rows",
        "short": "-r",
        "nargs": "?",
        "type": int,
        "default": 4,
        "help": "The number of rows of the board to generate"
    },
    {
        "long": "--cols",
        "short": "-c",
        "nargs": "?",
        "type": int,
        "default": 4,
        "help": "The number of columns of the board to generate"
    },
    {
        "long": "--sort",
        "short": "-s",
        "nargs": "?",
        "type": str,
        "default": "score",
        "help": "Sort words by [score|length|start|end]"
    },
    {
        "long": "--reverse",
        "short": "-R",
        "action": "store_true",
        "help": "Reverse the list of words"
    },
    {
        "long": "--quiet",
        "short": "-q",
        "action": "store_true",
        "help": "Do not output board and statistics"
    },
]

def print_error(msg):
    """
    Print the given error message and exit.

    :param str msg: the error message
    """
    print(msg)
    sys.exit(1)

def check_language(vargs):
    """
    Check that the language has been specified,
    and that it is among the supported ones.
    On error, print error message and exit. 

    :param dict vargs: the command line arguments
    """
    if vargs["language"] is None:
        print_error("You must specify the language of the board.")
    if vargs["language"] not in LANGUAGES:
        print_error("You must specify a supported language: %s" % ", ".join(LANGUAGES))

def list_languages():
    """
    List all the available languages
    """
    print("Supported languages: %s" % ", ".join(LANGUAGES))

def demo_mode(vargs):
    """
    Solve a built-in board.
    
    :param dict vargs: the command line arguments
    """
    check_language(vargs)
    vargs["board"] = os.path.abspath(os.path.join(os.path.dirname(__file__), "res/" + vargs["language"] + ".board"))
    solve_board(vargs)

def solve_board(vargs):
    """
    Solve a board.
    
    :param dict vargs: the command line arguments
    """
    check_language(vargs)
    if vargs["board"] is None:
        print_error("You must specify the path of the board file to solve.")
    if vargs["dictionary"] is None:
        vargs["dictionary"] = os.path.abspath(os.path.join(os.path.dirname(__file__), "res/" + vargs["language"] + ".marisa"))
    dictionary = MTDictionary(vargs["dictionary"], normalize=True, ignore_case=True)
    board = Board(vargs["language"]).read_board_file(vargs["board"])
    if not vargs["quiet"]:
        print("")
        print(board.pretty_print(multipliers=True))
        print("")
    solver = Solver(board, dictionary)
    words = solver.solve(sort=vargs["sort"], reverse=vargs["reverse"])
    length_longest_word = max([len(w[0]) for w in words])
    length_max_score = len(str(max([w[1] for w in words])))
    total = 0
    for (word, snake_score, snake) in words:
        word_padding = " " * (length_longest_word - len(word))
        score_padding = " " * (length_max_score - len(str(snake_score)))
        print("%s%s    %d%s    %s" % (word, word_padding, snake_score, score_padding, snake))
        total += snake_score
    if not vargs["quiet"]:
        print("")
        print("Number of words:            %d" % len(words))
        print("Length of the longest word: %d" % length_longest_word)
        print("Maximum total score:        %d" % total)
        print("")

def generate_board(vargs):
    """
    Generate a random board.
    
    :param dict vargs: the command line arguments
    """
    check_language(vargs)
    board = Board(vargs["language"]).generate_random_board(rows=vargs["rows"], cols=vargs["cols"])
    print(board.pretty_print(multipliers=True))
    if vargs["output"] is not None:
        board.save_to_file(vargs["output"])
        print("")
        print("File '%s' saved" % vargs["output"])

def cat_dictionary(vargs):
    """
    Output the keys in the given dictionary to stdout.
    No normalization will take place!
    
    :param dict vargs: the command line arguments
    """
    if vargs["dictionary"] is None:
        print_error("You must specify the path of the input dictionary file.")
    words = MTDictionary(vargs["dictionary"], normalize=False, ignore_case=False)
    if vargs["output"] is not None:
        words.save_plain_text(vargs["output"])
        print("File '%s' saved" % vargs["output"])
    else:
        for word in words.keys:
            print(word)

def compile_dictionary(vargs):
    """
    Compile the given dictionary into a MARISA trie,
    and save it to file.
    
    :param dict vargs: the command line arguments
    """
    if vargs["dictionary"] is None:
        print_error("You must specify the path of the input dictionary file.")
    if vargs["output"] is None:
        print_error("You must specify the path of the output dictionary file.")
    output_file_path = vargs["output"]
    if not output_file_path.endswith(".marisa"):
        output_file_path += ".marisa"
    words = MTDictionary(vargs["dictionary"], normalize=True, ignore_case=True)
    words.save_marisa_trie(output_file_path)
    print("File '%s' saved" % output_file_path)

def main():
    """
    Entry point.
    """
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    for arg in ARGUMENTS:
        if "action" in arg:
            if arg["short"] is not None:
                parser.add_argument(arg["short"], arg["long"], action=arg["action"], help=arg["help"])
            else:
                parser.add_argument(arg["long"], action=arg["action"], help=arg["help"])
        else:
            if arg["short"] is not None:
                parser.add_argument(arg["short"], arg["long"], nargs=arg["nargs"], type=arg["type"], default=arg["default"], help=arg["help"])
            else:
                parser.add_argument(arg["long"], nargs=arg["nargs"], type=arg["type"], default=arg["default"], help=arg["help"])
    vargs = vars(parser.parse_args())
    command = vargs["command"]
    if command == "languages":
        list_languages()
    elif command == "demo":
        demo_mode(vargs)
    elif command == "solve":
        solve_board(vargs)
    elif command == "generate":
        generate_board(vargs)
    elif command == "cat":
        cat_dictionary(vargs)
    elif command == "compile":
        compile_dictionary(vargs)
    else:
        parser.print_help()
        sys.exit(2)
    sys.exit(0)



if __name__ == "__main__":
    main()






#!/usr/bin/env python
# coding=utf-8

"""
A Ruzzle board.
"""

from __future__ import absolute_import
from __future__ import print_function
import io
import os
import random

from elzzur.languages import LANGUAGES, LETTER_SCORE, LETTER_FREQUENCY

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2016, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "0.0.1"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

TRIPLE_WORD = "tw"
""" Triple word multiplier placeholder """

DOUBLE_WORD = "dw"
""" Double word multiplier placeholder """

TRIPLE_LETTER = "tl"
""" Triple letter multiplier placeholder """

DOUBLE_LETTER = "dl"
""" Double letter multiplier placeholder """

LENGTH_POINTS = {
    9: 25,
    8: 20,
    7: 15,
    6: 10,
    5: 5,
}
""" Length points """

MULTIPLIERS = {
    TRIPLE_WORD: (3, 1),
    DOUBLE_WORD: (2, 1),
    TRIPLE_LETTER: (1, 3),
    DOUBLE_LETTER: (1, 2),
}
""" Values of the multipliers """

MULTIPLIERS_CDF = [
    (TRIPLE_WORD, 0.05),    #  5%
    (DOUBLE_WORD, 0.15),    # 10%
    (TRIPLE_LETTER, 0.25),  # 10%
    (DOUBLE_LETTER, 0.40),  # 15%
    ("", 1.0)               # 60%
]
""" CDF of the multipliers, used for generating random boards """

class BoardCell(object):
    """
    A cell of the board.
    
    :param str token: the token, containing the letter (e.g., ``E``) and possibly its multiplier (e.g., ``Mtw`` or ``Adl``)
    :param str language: the language code (e.g. ``en``) of the board, used to determine the letter score
    """
    def __init__(self, token, language):
        self.letter = token[0].upper()
        if self.letter not in LETTER_SCORE[language]:
            raise ValueError("Unrecognized letter '%s' for language '%s'." % (token[0], language))
        self.word_multiplier = 1
        self.letter_multiplier = 1
        self.token_multiplier = ""
        if len(token) > 1:
            self.token_multiplier = token[1:].lower()
            if self.token_multiplier not in MULTIPLIERS:
                raise ValueError("Unrecognized multiplier '%s'." % (token[1:]))
            self.word_multiplier, self.letter_multiplier = MULTIPLIERS[self.token_multiplier]
        self.score = LETTER_SCORE[language][self.letter] * self.letter_multiplier

    def __str__(self):
        return u"%s (L=%d, W=%d)" % (self.letter, self.score, self.word_multiplier)

    def pretty_print(self, multiplier=False):
        """
        Pretty print the cell.

        :param bool multiplier: if ``True``, print the multiplier
        """
        if multiplier:
            return u"%s%s" % (self.letter, self.token_multiplier) 
        return u"%s" % self.letter

class Board(object):
    """
    A Ruzzle board.

    :param str language: the language code (e.g. ``en``) of the board, used to determine the letter score
    """
    def __init__(self, language):
        if language not in LETTER_SCORE:
            raise ValueError("No score available for the given language. (Got '%s')" % language)
        self.language = language
        self.cells = {}
        self.rows = 0
        self.cols = 0

    def __str__(self):
        acc = []
        for i in range(self.rows):
            acc.append(u" ".join([str(self.cells[(i, j)]) for j in range(self.cols)]))
        return u"\n".join(acc)

    def pretty_print(self, multipliers=False):
        """
        Pretty print the board.

        :param bool multipliers: if ``True``, print the multipliers
        :rtype: str
        """
        acc = []
        for i in range(self.rows):
            row = ""
            for j in range(self.cols):
                letter = self.cells[(i, j)].pretty_print(multiplier=multipliers)
                padding = " " * (4 - len(letter))
                row += letter + padding
            acc.append(row)
        return u"\n".join(acc)

    @property
    def letters(self):
        """
        Return the list of letters in the board.

        :rtype: list of str
        """
        return [l.letter for l in list(self.cells.values())]

    def letter_at(self, cell):
        """
        Return the letter at the given cell.
        
        :param tuple cell: the ``(x, y)`` cell to be checked for
        """
        return self.cells[cell].letter

    def compute_snake_word(self, snake):
        """
        Return the word corresponding to the given snake.

        :param Snake snake: the snake corresponding to the desired word
        :rtype: str
        """
        return u"".join([self.cells[cell].letter for cell in snake.cells])

    def compute_snake_score(self, snake):
        """
        Return the score corresponding to the given snake.

        :param Snake snake: the snake corresponding to the desired word
        :rtype: int
        """
        acc = sum([self.cells[cell].score for cell in snake.cells])
        for cell in snake.cells:
            acc *= self.cells[cell].word_multiplier
        if len(snake) in LENGTH_POINTS:
            acc += LENGTH_POINTS[len(snake)]
        return acc

    def save_to_file(self, file_path):
        """
        Save the board to file.
        
        :param str file_path: the path of the output file to be written
        """
        with io.open(file_path, "w", encoding="utf-8") as f:
            f.write(self.pretty_print(multipliers=True))

    def read_board_file(self, file_path):
        """
        Read a plain text, UTF-8 encoded file,
        containing one line per board row,
        with letters separated by one or more spaces.

        Example::

            Ttl R   S Ndl
            Odw Htw E I
            Cdw I   N V
            Etl A   D E


        :param str file_path: the path of the input file to be read
        :rtype: Board
        """
        if not os.path.isfile(file_path):
            raise IOError("The board file does not exist. (Got: '%s')" % file_path)
        self.cells = {}
        row = 0
        cols = []
        with io.open(file_path, "r", encoding="utf-8") as f:
            for line in f:
                if len(line.strip()) == 0:
                    break
                col = 0
                for t in [l for l in line.strip().split(" ") if len(l) > 0]:
                    self.cells[(row, col)] = BoardCell(t, self.language)
                    col += 1
                cols.append(col)
                row += 1
        # check that we have at least one row
        if row < 1:
            raise ValueError("The board file seems empty.")
        # check that all rows have the same number of columns
        for j in range(len(cols) - 1):
            if cols[j] != cols[j+1]:
                raise ValueError("The board file has two rows with a different number of columns. (Got: row %d has %d cols while row %d had %d cols))" % (j, cols[j], j+1, cols[j+1]))
        # all good, store the number of rows and columns
        self.rows = row
        self.cols = cols[0]
        return self

    def generate_random_board(self, rows=4, cols=4):
        """
        Generate a random board.

        :param int rows: the number of rows of the desired board
        :param int cols: the number of columns of the desired board
        :rtype: Board
        """
        def cumsum(letter_freq):
            # numpy has cumsum, but I do not want numpy as dependency
            cum_freq = []
            f_prev = 0.0
            for (l, f) in letter_freq:
                f_prev += f
                cum_freq.append((l, f_prev))
            f_sum = cum_freq[-1][1]
            return [(l, f/f_sum) for (l, f) in cum_freq]

        def draw_from_distribution(cdf):
            #
            # numpy has digitize() and numpy.random, but I do not want numpy as dependency
            # this can be done with binary search instead of linear search
            #
            # in general, one should use the distribution of N-grams, N>=2 instead of single letter frequencies (N=1)
            # I suspect that the original Ruzzle builds the board so that there are at least M words, etc.
            # (i.e., the boards are not truly "random", based on some N-gram distribution)
            #
            rnd = random.uniform(0, 1)
            for (element, cf) in cdf:
                if rnd <= cf:
                    return element

        self.cells = {}
        self.rows = rows
        self.cols = cols
        sorted_letters = sorted([(l, LETTER_FREQUENCY[self.language][l]) for l in LETTER_FREQUENCY[self.language]])
        cum_freq = cumsum(sorted_letters)
        for row in range(self.rows):
            for col in range(self.cols):
                letter = draw_from_distribution(cum_freq)
                multiplier = draw_from_distribution(MULTIPLIERS_CDF)
                self.cells[(row, col)] = BoardCell(letter + multiplier, self.language)
        return self






#!/usr/bin/env python
# coding=utf-8

"""
Solve a Ruzzle board.

The idea is to do a BFS exploration of the board,
avoiding extending the current snake (i.e., adjacent sequence of letters),
if either
a. the snake self-intersects or,
b. the word corresponding to the current snake is not a prefix of any word in the dictionary.

To speed the lookup operations, the dictionary is stored in a MARISA trie,
which is a very efficient trie (a.k.a. prefix tree),
supporting the has_keys_with_prefix(prefix) operation.
"""

from __future__ import absolute_import
from __future__ import print_function

from elzzur.board import Board
from elzzur.mtdictionary import MTDictionary
from elzzur.snake import Snake

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2016, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "0.0.1"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

class Solver(object):
    """
    Solve a Ruzzle board.

    Please see the README.md for a discussion.

    :param Board board: the board to solve
    :param MTDictionary dictionary: the dictionary containing the valid words
    """

    SORT_BY_SCORE = "score"
    """ Sort by score (decr), length (inc), word (alpha) """

    SORT_BY_LENGTH = "length"
    """ Sort by length (decr), score (decr), word (alpha) """

    SORT_BY_START = "start"
    """ Sort by snake start position (NW->SE), score (decr), word (alpha) """

    SORT_BY_END = "end"
    """ Sort by snake end position (NW->SE), score (decr), word (alpha) """

    def __init__(self, board, dictionary):
        self.board = board
        self.dictionary = dictionary
        self.found = {}

    def solve(self, sort=SORT_BY_SCORE, reverse=False):
        """
        Solve the board.

        :param str sort: the sort method
        :param bool reverse: if ``True`` reverse the order of the words
        """
        self.found = {}
        # find all valid snakes
        for snake in self.find_snakes():
            # for each word, keep only the snake with the highest score
            snake_word = self.board.compute_snake_word(snake)
            snake_score = self.board.compute_snake_score(snake)
            if (snake_word not in self.found) or (self.found[snake_word][1] < snake_score):
                self.found[snake_word] = (snake_word, snake_score, snake)
        # sort and return
        return self.sort_words(sort=sort, reverse=reverse)

    def find_snakes(self):
        """
        Find all the valid snakes in the board.

        :rtype: list of Snake objects
        """
        valid_snakes = []
        rows = self.board.rows
        cols = self.board.cols
        for row in range(rows):
            for col in range(cols):
                to_be_explored = [Snake([(row, col)])]
                while len(to_be_explored) > 0:
                    current = to_be_explored.pop(0)
                    current_word = self.board.compute_snake_word(current)
                    if (len(current_word) > 1) and (self.dictionary.has_key(current_word)):
                        valid_snakes.append(current)
                    if self.dictionary.has_keys_with_prefix(current_word):
                        crow, ccol = current.end
                        for trow in range(max(0, crow - 1), min(rows, crow + 2)):       # (crow + 1) + 1 => range(x, y) = [x, x+1, ... , y-1]
                            for tcol in range(max(0, ccol - 1), min(cols, ccol + 2)):   # (ccol + 1) + 1 => range(x, y) = [x, x+1, ... , y-1]
                                tcell = (trow, tcol) 
                                if not current.has_cell(tcell):
                                    to_be_explored.append(current.extend(tcell))
        return valid_snakes
            
    def sort_words(self, sort=SORT_BY_SCORE, reverse=False):
        """
        Sort the found words according to the requested method,
        and return them as a list.

        :param str sort: the sort method
        :param bool reverse: if ``True`` reverse the order of the words
        :rtype: list of
        """
        if sort == self.SORT_BY_LENGTH:
            key = lambda x: (len(x[0]), x[1], x[0])
            rev = True
        elif sort == self.SORT_BY_START:
            key = lambda x: (x[2].start, x[1], x[0])
            rev = False
        elif sort == self.SORT_BY_END:
            key = lambda x: (x[2].end, x[1], x[0])
            rev = False
        else:
            key = lambda x: (x[1], -len(x[0]), x[0])
            rev = True
        reverse = not rev if reverse else rev
        return sorted([self.found[word] for word in self.found], key=key, reverse=reverse)






#!/usr/bin/env python
# coding=utf-8

"""
A dictionary based on MARISA trie.

This implementation requires the ``marisa-trie`` Python package (``pip install marisa-trie``).
"""

from __future__ import absolute_import
from __future__ import print_function
import io
import marisa_trie
import os
import unicodedata

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2016, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "0.0.1"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

class MTDictionary(object):
    """
    A dictionary based on a MARISA trie.

    A trie (a.k.a. prefix tree) dictionary allows for very fast checking
    if there are words in the dictionary starting with a given prefix,
    and, if so, to retrieve them all.

    MARISA is a very efficient trie implementation.

    This class allows reading a dictionary from
    a. a MARISA file, or
    b. a plain text, UTF-8 encoded file.
    In the latter case, you can save the resulting MARISA trie to file to use it later.

    :param str dictionary_file_path: path to the dictionary file to read. If it ends with ``.marisa``, it is read as a MARISA trie.
    :param bool normalize: if ``True``, apply Unicode NFKD + decode to ascii to the dictionary entries
    :param bool ignore_case: if ``True``, ignore case, that is, make all dictionary entries uppercase
    """
    def __init__(self, dictionary_file_path, normalize=False, ignore_case=False):
        if not os.path.isfile(dictionary_file_path):
            raise IOError("The dictionary file does not exist. (Got: '%s')" % dictionary_file_path)
        if dictionary_file_path.endswith(".marisa"):
            self.read_marisa_file(dictionary_file_path)
        else:
            self.read_plain_file(dictionary_file_path, normalize=normalize, ignore_case=ignore_case)

    def __len__(self):
        return len(self.trie)

    @property
    def keys(self):
        """
        Return the sorted list of keys in the dictionary.

        :rtype: list of str
        """
        return sorted([w for w in self.trie])

    def has_key(self, key):
        """
        Return ``True`` if the given key is present in the dictionary.

        :param str key: the key (word) to be checked for
        :rtype: bool
        """
        return key in self.trie

    def has_keys_with_prefix(self, prefix):
        """
        Return ``True`` if in the dictionary there are keys with the given prefix.

        :param str key: the prefix (word prefix) to be checked for
        :rtype: bool
        """
        return self.trie.has_keys_with_prefix(prefix)

    def keys_with_prefix(self, prefix):
        """
        Return a list of keys in the dictionary with the given prefix.

        :param str key: the prefix (word prefix) to be checked for
        :rtype: list of str
        """
        return self.trie.keys(prefix)

    def read_marisa_file(self, file_path):
        """
        Read a MARISA trie from file and return it.

        :param str file_path: the path of the input file to be read
        """
        self.trie = marisa_trie.Trie()
        with io.open(file_path, "rb") as f:
            self.trie.read(f)

    def read_plain_file(self, file_path, normalize=False, ignore_case=False):
        """
        Read a plain text, UTF-8 encoded file,
        containing one word per line,
        and return a MARISA trie from it.
        
        :param str file_path: the path of the input file to be read
        :param bool normalize: if ``True``, apply Unicode NFKD + decode to ascii to the dictionary entries
        :param bool ignore_case: if ``True``, ignore case, that is, make all dictionary entries uppercase
        """
        with io.open(file_path, "r", encoding="utf-8") as f:
            dictionary = f.read()
        if normalize:
            dictionary = unicodedata.normalize("NFKD", dictionary).encode("ascii", "ignore")
        if ignore_case:
            dictionary = dictionary.upper()
        words = []
        for line in dictionary.split(u"\n"):
            words.append(line.strip())
        self.trie = marisa_trie.Trie(words)

    def save_marisa_trie(self, file_path):
        """
        Save the MARISA trie to file.
        
        :param str file_path: the path of the output file to be written
        """
        with io.open(file_path, "wb") as f:
            self.trie.write(f)

    def save_plain_text(self, file_path):
        """
        Save the dictionary to file as plain text.
        
        :param str file_path: the path of the output file to be written
        """
        with io.open(file_path, "w", encoding="utf-8") as f:
            f.write(u"\n".join(self.keys))






#!/usr/bin/env python
# coding=utf-8

"""
elzzur solves a Ruzzle board, listing all the valid words with their scores.
"""

from __future__ import absolute_import
from __future__ import print_function

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2016, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "0.0.1"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

from elzzur.__main__ import main






#!/usr/bin/env python
# coding=utf-8

"""
Set elzzur package up.
"""

from setuptools import setup, Extension

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2016, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "0.0.1"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

setup(
    name="elzzur",
    packages=["elzzur"],
    package_data={"elzzur": ["res/*"]},
    version="0.0.1.0",
    description="elzzur solves a Ruzzle board, listing all the valid words with their scores.",
    author="Alberto Pettarin",
    author_email="alberto@albertopettarin.it",
    url="https://github.com/pettarin/elzzur",
    license="MIT License",
    long_description=open("README.rst", "r").read(),
    install_requires=["marisa-trie>=0.7.2"],
    scripts=["bin/elzzur"],
    keywords=[
        "elzzur",
        "Ruzzle",
        "Ramble",
        "Scrabble",
        "Scarabeo",
        "Paroliere",
        "Snake",
        "board",
        "snake",
        "snakes",
        "solve",
        "solver",
        "MARISA",
        "trie",
        "prefix tree"
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Games/Entertainment",
        "Topic :: Games/Entertainment :: Board Games",
        "Topic :: Games/Entertainment :: Puzzle Games",
        "Topic :: Desktop Environment",
        "Topic :: Utilities"
    ],
)



#!/usr/bin/env python
# coding=utf-8

"""
A Snake represents a (possibly partial) list of adjacent board cells,
represented by ``(x, y)`` pairs, 0-indexing.
"""

from __future__ import absolute_import
from __future__ import print_function

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2016, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "0.0.1"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

class Snake(object):
    """
    A Snake represents a (possibly partial) list of adjacent board cells,
    represented by ``(x, y)`` pairs, 0-indexing.
    
    :param list cells: list of (x, y) pairs representing the cells of the snake
    """
    def __init__(self, cells):
        self.cells = cells

    def __str__(self):
        return u" ".join(["(%d,%d)" % cell for cell in self.cells])

    def __len__(self):
        return len(self.cells)

    @property
    def start(self):
        """
        The cell where the snake starts.

        :rtype: (int, int)
        """
        return self.cells[0]

    @property
    def end(self):
        """
        The cell where the snake ends.

        :rtype: (int, int)
        """
        return self.cells[-1]
    
    def has_cell(self, cell):
        """
        Return ``True`` if the given ``(x, y)`` cell is already in the snake.

        :param tuple cell: the ``(x, y)`` cell to be checked for
        :rtype: bool
        """
        return cell in self.cells
    
    def extend(self, cell):
        """
        Return a new Snake which is the current Snake extended with the given cell.

        :param tuple cell: the ``(x, y)`` cell to be added
        :rtype: Snake
        """
        return Snake(self.cells + [cell])



